/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.token;

import java.util.Hashtable;

/**
 * ȡΥפ̤ޤ
 *
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.3 $
 */
public final class TokenTypes {
    /**                */ public static final int UNKNOWN          =   -1;
    /** ƥ           */ public static final int TEXT             =    2;
    /** DTD                */ public static final int DTD              =    3;
    /**            */ public static final int COMMENT          =    4;
    /** ̿           */ public static final int PI               =    5;
    /** ¾             */ public static final int MISC             =    0;

    /** ʸ񳫻           */ public static final int HTML_START       = 1001;
    /** ʸλ           */ public static final int HTML_END         = 1002;

    /** إå       */ public static final int HEAD_START       = 2001;
    /** إåλ       */ public static final int HEAD_END         = 2002;

    /** ȥ볫       */ public static final int TITLE_START      = 2011;
    /** ȥ뽪λ       */ public static final int TITLE_END        = 2012;
    /** ١             */ public static final int BASE             = 2013;
    /** ץȳ     */ public static final int SCRIPT_START     = 2015;
    /** ץȽλ     */ public static final int SCRIPT_END       = 2016;
    /** 볫       */ public static final int STYLE_START      = 2017;
    /** 뽪λ       */ public static final int STYLE_END        = 2018;
    /** ᥿           */ public static final int META             = 2019;
    /** 󥯾         */ public static final int LINK             = 2021;

    private static final int FRAME_TYPE_START                          = 2041;
    /** ե졼ॻåȳ */ public static final int FRAMESET_START   = 2041;
    /** ե졼ॻåȽλ */ public static final int FRAMESET_END     = 2042;
    /** ե졼           */ public static final int FRAME            = 2043;
    /** ե졼̵   */ public static final int NOFRAMES_START   = 2045;
    /** ե졼̵λ   */ public static final int NOFRAMES_END     = 2046;
    private static final int FRAME_TYPE_END                            = 2050;

    private static final int BODY_TYPE_START                           = 3001;
    /** ʸ           */ public static final int BODY_START       = 3001;
    /** ʸλ           */ public static final int BODY_END         = 3002;

    //### block
    private static final int BLOCK_TYPE_START                          = 3101;
    /**            */ public static final int P_START          = 3101;
    /** λ           */ public static final int P_END            = 3102;
    /** Ф1        */ public static final int H1_START         = 3103;
    /** Ф1λ        */ public static final int H1_END           = 3104;
    /** Ф2        */ public static final int H2_START         = 3105;
    /** Ф2λ        */ public static final int H2_END           = 3106;
    /** Ф3        */ public static final int H3_START         = 3107;
    /** Ф3λ        */ public static final int H3_END           = 3108;
    /** Ф4        */ public static final int H4_START         = 3109;
    /** Ф4λ        */ public static final int H4_END           = 3110;
    /** Ф5        */ public static final int H5_START         = 3111;
    /** Ф5λ        */ public static final int H5_END           = 3112;
    /** Ф6        */ public static final int H6_START         = 3113;
    /** Ф6λ        */ public static final int H6_END           = 3114;
    /** վ񤭳       */ public static final int UL_START         = 3115;
    /** վ񤭽λ       */ public static final int UL_END           = 3116;
    /** ꥹȳ     */ public static final int OL_START         = 3117;
    /** ꥹȽλ     */ public static final int OL_END           = 3118;
    /** ǥ쥯ȥ곫   */ public static final int DIR_START        = 3119;
    /** ǥ쥯ȥ꽪λ   */ public static final int DIR_END          = 3120;
    /** ˥塼       */ public static final int MENU_START       = 3121;
    /** ˥塼λ       */ public static final int MENU_END         = 3122;
    /** ꥹȹܳ     */ public static final int LI_START         = 3123;
    /** ꥹȹܽλ     */ public static final int LI_END           = 3124;
    /** ꥹȳ     */ public static final int DL_START         = 3125;
    /** ꥹȽλ     */ public static final int DL_END           = 3126;
    /** 쳫         */ public static final int DT_START         = 3127;
    /** 콪λ         */ public static final int DT_END           = 3128;
    /**      */ public static final int DD_START         = 3129;
    /** λ     */ public static final int DD_END           = 3130;
    /** Ѥ   */ public static final int PRE_START        = 3131;
    /** Ѥλ   */ public static final int PRE_END          = 3132;
    /** ֥å       */ public static final int DIV_START        = 3133;
    /** ֥åλ       */ public static final int DIV_END          = 3134;
    // NOSCRIPT ϥ饤Ͽ
    /** ʸ         */ public static final int BLOCKQUOTE_START = 3137;
    /** ʸλ         */ public static final int BLOCKQUOTE_END   = 3138;
    /**                  */ public static final int HR               = 3139;
    /** ɥ쥹       */ public static final int ADDRESS_START    = 3141;
    /** ɥ쥹λ       */ public static final int ADDRESS_END      = 3142;
    /** 󥿥󥰳   */ public static final int CENTER_START     = 3143;      // 侩
    /** 󥿥󥰽λ   */ public static final int CENTER_END       = 3144;      // 侩
    /** ۥ֥å   */ public static final int X_BLOCK_START    = 3191;
    /** ۥ֥åλ   */ public static final int X_BLOCK_END      = 3192;
    private static final int BLOCK_TYPE_END                            = 3200;

    //### inline
    private static final int INLINE_TYPE_START                         = 3201;
    /** 󥫡       */ public static final int A_START          = 3201;
    /** 󥫡λ       */ public static final int A_END            = 3202;
    /** ᡼           */ public static final int IMG              = 3203;
    /**                */ public static final int BR               = 3205;
    /** SPAN           */ public static final int SPAN_START       = 3207;
    /** SPAN λ          */ public static final int SPAN_END         = 3208;
    /** EM             */ public static final int EM_START         = 3221;
    /** EM λ            */ public static final int EM_END           = 3222;
    /** STRONG         */ public static final int STRONG_START     = 3223;
    /** STRONG λ        */ public static final int STRONG_END       = 3224;
    /** DFN            */ public static final int DFN_START        = 3225;
    /** DFN λ           */ public static final int DFN_END          = 3226;
    /** CODE           */ public static final int CODE_START       = 3227;
    /** CODE λ          */ public static final int CODE_END         = 3228;
    /** Q              */ public static final int Q_START          = 3229;
    /** Q λ             */ public static final int Q_END            = 3230;
    /** SUB            */ public static final int SUB_START        = 3231;
    /** SUB λ           */ public static final int SUB_END          = 3232;
    /** SUP            */ public static final int SUP_START        = 3233;
    /** SUP λ           */ public static final int SUP_END          = 3234;
    /** SAMP           */ public static final int SAMP_START       = 3235;
    /** SAMP λ          */ public static final int SAMP_END         = 3236;
    /** KBD            */ public static final int KBD_START        = 3237;
    /** KBD λ           */ public static final int KBD_END          = 3238;
    /** VAR            */ public static final int VAR_START        = 3239;
    /** VAR λ           */ public static final int VAR_END          = 3240;
    /** CITE           */ public static final int CITE_START       = 3241;
    /** CITE λ          */ public static final int CITE_END         = 3242;
    /** ABBR           */ public static final int ABBR_START       = 3243;
    /** ABBR λ          */ public static final int ABBR_END         = 3244;
    /** ACRONYM        */ public static final int ACRONYM_START    = 3245;
    /** ACRONYM λ       */ public static final int ACRONYM_END      = 3246;
    /** TT             */ public static final int TT_START         = 3261;
    /** TT λ            */ public static final int TT_END           = 3262;
    /** I              */ public static final int I_START          = 3263;
    /** I λ             */ public static final int I_END            = 3264;
    /** B              */ public static final int B_START          = 3265;
    /** B λ             */ public static final int B_END            = 3266;
    /** BIG            */ public static final int BIG_START        = 3267;
    /** BIG λ           */ public static final int BIG_END          = 3268;
    /** SMALL          */ public static final int SMALL_START      = 3269;
    /** SMALL λ         */ public static final int SMALL_END        = 3270;
    /** U              */ public static final int U_START          = 3271;    // 侩
    /** U λ             */ public static final int U_END            = 3272;    // 侩
    /** S              */ public static final int S_START          = 3273;    // 侩
    /** S λ             */ public static final int S_END            = 3274;    // 侩
    /** STRIKE         */ public static final int STRIKE_START     = 3275;    // 侩
    /** STRIKE λ        */ public static final int STRIKE_END       = 3276;    // 侩
    /** FONT           */ public static final int FONT_START       = 3277;    // 侩
    /** FONT λ          */ public static final int FONT_END         = 3278;    // 侩
    /** 󥹥ץȳ   */ public static final int NOSCRIPT_START   = 3291;    // ϥ֥å
    /** 󥹥ץȽλ   */ public static final int NOSCRIPT_END     = 3292;    // ϥ֥å
    private static final int INLINE_TYPE_END                           = 3300;

    //### ruby
    private static final int RUBY_TYPE_START                           = 3301;
    /** RUBY           */ public static final int RUBY_START       = 3301;
    /** RUBY λ          */ public static final int RUBY_END         = 3302;
    /** RB             */ public static final int RB_START         = 3303;
    /** RB λ            */ public static final int RB_END           = 3304;
    /** RP             */ public static final int RP_START         = 3305;
    /** RP λ            */ public static final int RP_END           = 3306;
    /** RT             */ public static final int RT_START         = 3307;
    /** RT λ            */ public static final int RT_END           = 3308;
    private static final int RUBY_TYPE_END                             = 3310;

    //### table
    private static final int TABLE_TYPE_START                          = 3401;
    /** TABLE          */ public static final int TABLE_START      = 3401;
    /** TABLE λ         */ public static final int TABLE_END        = 3402;
    /** CAPTION        */ public static final int CAPTION_START    = 3403;
    /** CAPTION λ       */ public static final int CAPTION_END      = 3404;
    /** THEAD          */ public static final int THEAD_START      = 3411;
    /** THEAD λ         */ public static final int THEAD_END        = 3412;
    /** TBODY          */ public static final int TBODY_START      = 3413;
    /** TBODY λ         */ public static final int TBODY_END        = 3414;
    /** TFOOT          */ public static final int TFOOT_START      = 3415;
    /** TFOOT λ         */ public static final int TFOOT_END        = 3416;
    /** TR             */ public static final int TR_START         = 3421;
    /** TR λ            */ public static final int TR_END           = 3422;
    /** TH             */ public static final int TH_START         = 3431;
    /** TH λ            */ public static final int TH_END           = 3432;
    /** TD             */ public static final int TD_START         = 3433;
    /** TD λ            */ public static final int TD_END           = 3434;
    private static final int TABLE_TYPE_END                            = 3450;

    //### form
    private static final int FORM_TYPE_START                           = 3501;
    /** FORM           */ public static final int FORM_START       = 3501;
    /** FORM λ          */ public static final int FORM_END         = 3502;
    /** INPUT          */ public static final int INPUT            = 3503;
    /** SELECT         */ public static final int SELECT_START     = 3505;
    /** SELECT λ        */ public static final int SELECT_END       = 3506;
    /** OPTGROUP       */ public static final int OPTGROUP_START   = 3507;
    /** OPTGROUP λ      */ public static final int OPTGROUP_END     = 3508;
    /** OPTION         */ public static final int OPTION_START     = 3509;
    /** OPTION λ        */ public static final int OPTION_END       = 3510;
    /** TEXTAREA       */ public static final int TEXTAREA_START   = 3511;
    /** TEXTAREA λ      */ public static final int TEXTAREA_END     = 3512;
    /** BUTTON         */ public static final int BUTTON_START     = 3513;
    /** BUTTON λ        */ public static final int BUTTON_END       = 3514;
    /** FIELDSET       */ public static final int FIELDSET_START   = 3515;
    /** FIELDSET λ      */ public static final int FIELDSET_END     = 3516;
    /** LEGEND         */ public static final int LEGEND_START     = 3517;
    /** LEGEND λ        */ public static final int LEGEND_END       = 3518;
    /** LABEL          */ public static final int LABEL_START      = 3519;
    /** LABEL λ         */ public static final int LABEL_END        = 3520;
    private static final int FORM_TYPE_END                             = 3550;

    private static final int BODY_TYPE_END                             = 4000;

    private static final Hashtable TOKEN_NAMES = new Hashtable();

    private static final String[] TITLE_TRAP   = {"base", "script", "style", "meta", "link", "head", "body"};
    private static final String[] OPTION_TRAP  = {"select"};

    /** ߡ */
    private TokenTypes() {
    }

    /**
     * ꤵ줿פ BODY ʬξ <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return BODY ʬξ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isBody(int type) {
        return (BODY_TYPE_START <= type && type <= BODY_TYPE_END);
    }

    /**
     * ꤵ줿פ֥åǤξ <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return ֥åǤξ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isBlock(int type) {
        if (BLOCK_TYPE_START <= type && type <= BLOCK_TYPE_END) {
            return true;
        }
        switch (type) {
        case TABLE_START   :
        case TABLE_END     :
        case FORM_START    :
        case FORM_END      :
        case FIELDSET_START:
        case FIELDSET_END  :
            return true;
        default: // AVOID
        }
        return false;
    }

    /**
     * ꤵ줿פΥ֥åǤȤư
     * <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return Υ֥åǤȤư <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isBlockEx(int type) {
        if (BLOCK_TYPE_START <= type && type <= BLOCK_TYPE_END) {
            return true;
        }
        switch (type) {
        case TABLE_START   :
        case TABLE_END     :
        case CAPTION_START :
        case CAPTION_END   :
        case TH_START      :
        case TH_END        :
        case TD_START      :
        case TD_END        :
        case BODY_START    :
        case BODY_END      :
        case FORM_START    :
        case FORM_END      :
        case FIELDSET_START:
        case FIELDSET_END  :
            return true;
        default: // AVOID
        }
        return false;
    }

    /**
     * ꤵ줿פ饤Ǥξ <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return 饤Ǥξ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isInline(int type) {
        return ((INLINE_TYPE_START <= type && type <= INLINE_TYPE_END)
             || type == RUBY_START || type == RUBY_END);
    }

    /**
     * ꤵ줿פǤξ <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return Ǥξ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isRuby(int type) {
        return (RUBY_TYPE_START <= type && type <= RUBY_TYPE_END);
    }

    /**
     * ꤵ줿פӤǤξ <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return ӤǤξ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isRubyItem(int type) {
        return (RUBY_TYPE_START <= type && type <= RUBY_TYPE_END
             && type != RUBY_START && type != RUBY_END);
    }

    /**
     * ꤵ줿פơ֥Ǥξ <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return ơ֥Ǥξ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isTable(int type) {
        return (TABLE_TYPE_START <= type && type <= TABLE_TYPE_END);
    }

    /**
     * ꤵ줿פơ֥Ǥξ <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return ơ֥Ǥξ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isTableItem(int type) {
        return (TABLE_TYPE_START <= type && type <= TABLE_TYPE_END
             && type != TABLE_START && type != TABLE_END);
    }

    /**
     * ꤵ줿פեǤξ <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return եǤξ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isForm(int type) {
        return (FORM_TYPE_START <= type && type <= FORM_TYPE_END);
    }

    /**
     * ꤵ줿פեǤξ <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return եǤξ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isFormItem(int type) {
        return (FORM_TYPE_START <= type && type <= FORM_TYPE_END
             && type != FORM_START && type != FORM_END);
    }

    /**
     * ꤵ줿פե졼Ǥξ <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return ե졼Ǥξ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isFrame(int type) {
        return (FRAME_TYPE_START <= type && type <= FRAME_TYPE_END);
    }

    /**
     * ꤵ줿פǤξ <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return Ǥξ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean isEmpty(int type) {
        switch (type) {
        // header
        case BASE :
        case META :
        case LINK :
        // frame
        case FRAME:
        // block
        case HR   :
        // inline
        case IMG  :
        case BR   :
        // form
        case INPUT:
            return true;
        default: // AVOID
        }

        return false;
    }

    /**
     * Ȥ᤻ˡȡȤȤ¸뤫֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return ȡȤȤ¸ <code>true</code>
     *         ̾ξ <code>false</code>
     */
    public static boolean isContent(int type) {
        switch (type) {
        case SCRIPT_START  :
        case STYLE_START   :
        case TITLE_START   :
        case TEXTAREA_START:
        case OPTION_START  :
            return true;
        default: // AVOID
        }

        return false;
    }

    /**
     * Ȥ᤻ˡȡȤȤ¸ˡ
     * ȡȤˤϴޤޤʤȡ֤̾ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return <code>null</code> ξбĤޤǤȤȤȡ
     *         ʳξϡȤΥȡϤ
     *         ȡνȤʤȡ֤
     */
    public static String[] getTraps(int type) {
        switch (type) {
        case TITLE_START   :
            return TITLE_TRAP;
        case OPTION_START  :
            return OPTION_TRAP;
        default: // AVOID
        }

        return null;
    }

    /**
     * ꤵ줿פƥȥȡޤळȤ
     * <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return ޤळȤǽʾ <code>true</code>
     *         Ǥʤ <code>false</code>
     */
    public static boolean canHaveText(int type) {
        switch (type) {
        case HTML_START      : return false;
        // header
        case HEAD_START      : return false;

        case TITLE_START     : return true;
        case BASE            : return false;
        case SCRIPT_START    : return true;
        case STYLE_START     : return true;
        case META            : return false;
        case LINK            : return false;
        // frame
        case FRAMESET_START  : return false;
        case FRAME           : return false;
        case NOFRAMES_START  : return true;
        // body
        case BODY_START      : return false;
        // block
        case P_START         : return true;
        case H1_START        : return true;
        case H2_START        : return true;
        case H3_START        : return true;
        case H4_START        : return true;
        case H5_START        : return true;
        case H6_START        : return true;
        case UL_START        : return false;
        case OL_START        : return false;
        case DIR_START       : return false;
        case MENU_START      : return false;
        case LI_START        : return true;
        case DL_START        : return false;
        case DT_START        : return true;
        case DD_START        : return true;
        case PRE_START       : return true;
        case DIV_START       : return true;
        case BLOCKQUOTE_START: return false;
        case HR              : return false;
        case ADDRESS_START   : return true;
        case CENTER_START    : return true;
        case X_BLOCK_START   : return true;
        // inline
        case A_START         : return true;
        case IMG             : return false;
        case BR              : return false;
        case SPAN_START      : return true;
        case EM_START        : return true;
        case STRONG_START    : return true;
        case DFN_START       : return true;
        case CODE_START      : return true;
        case Q_START         : return true;
        case SUB_START       : return true;
        case SUP_START       : return true;
        case SAMP_START      : return true;
        case KBD_START       : return true;
        case VAR_START       : return true;
        case CITE_START      : return true;
        case ABBR_START      : return true;
        case ACRONYM_START   : return true;
        case TT_START        : return true;
        case I_START         : return true;
        case B_START         : return true;
        case BIG_START       : return true;
        case SMALL_START     : return true;
        case U_START         : return true;
        case S_START         : return true;
        case STRIKE_START    : return true;
        case FONT_START      : return true;
        case NOSCRIPT_START  : return true;
        // ruby
        case RUBY_START      : return false;
        case RB_START        : return true;
        case RP_START        : return true;
        case RT_START        : return true;
        // table
        case TABLE_START     : return false;
        case CAPTION_START   : return true;
        case THEAD_START     : return false;
        case TBODY_START     : return false;
        case TFOOT_START     : return false;
        case TR_START        : return false;
        case TH_START        : return true;
        case TD_START        : return true;
        // form
        case FORM_START      : return true;
        case INPUT           : return false;
        case SELECT_START    : return false;
        case OPTGROUP_START  : return false;
        case OPTION_START    : return true;
        case TEXTAREA_START  : return true;
        case BUTTON_START    : return true;
        case FIELDSET_START  : return true;
        case LEGEND_START    : return true;
        case LABEL_START     : return true;
        default: // AVOID
        }
        return false;
    }

    /**
     * ꤵ줿ˡ֥åǤƤ
     * <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return ֥åǤƤ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean canHaveBlock(int type) {
        switch (type) {
        case NOFRAMES_START  :
        case NOFRAMES_END    :
        case BODY_START      :
        case BODY_END        :
        // block
        case LI_START        :
        case LI_END          :
        case DD_START        :
        case DD_END          :
        case DIV_START       :
        case DIV_END         :
        case BLOCKQUOTE_START:
        case BLOCKQUOTE_END  :
        case CENTER_START    :
        case CENTER_END      :
        case NOSCRIPT_START  :
        case NOSCRIPT_END    :
        // table
        case TH_START        :
        case TH_END          :
        case TD_START        :
        case TD_END          :
            return true;
        default: // AVOID
        }
        return false;
    }

    /**
     * ꤵ줿פޡǤ
     * <code>true</code> ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return ޡǤ <code>true</code>
     *         ʳξ <code>false</code>
     */
    public static boolean canSetMargin(int type) {
        switch (type) {
        // table
        case TH_START:
        case TH_END  :
        case TD_START:
        case TD_END  :
            return false;
        default: // AVOID
        }
        return true;
    }

    /**
     * ꤵ줿פʸɽ֤ޤ
     *
     * @param  type ȡ󥿥
     *
     * @return ʸɽ
     *         ʾ <code>null</code>
     */
    public static String getName(int type) {
        switch (type) {
        case HTML_START      : return "html"      ;
        case HTML_END        : return "html"      ;
        // header
        case HEAD_START      : return "head"      ;
        case HEAD_END        : return "head"      ;

        case TITLE_START     : return "title"     ;
        case TITLE_END       : return "title"     ;
        case BASE            : return "base"      ;
        case SCRIPT_START    : return "script"    ;
        case SCRIPT_END      : return "script"    ;
        case STYLE_START     : return "style"     ;
        case STYLE_END       : return "style"     ;
        case META            : return "meta"      ;
        case LINK            : return "link"      ;
        // frame
        case FRAMESET_START  : return "frameset"  ;
        case FRAMESET_END    : return "frameset"  ;
        case FRAME           : return "frame"     ;
        case NOFRAMES_START  : return "noframes"  ;
        case NOFRAMES_END    : return "noframes"  ;
        // body
        case BODY_START      : return "body"      ;
        case BODY_END        : return "body"      ;
        // block
        case P_START         : return "p"         ;
        case P_END           : return "p"         ;
        case H1_START        : return "h1"        ;
        case H1_END          : return "h1"        ;
        case H2_START        : return "h2"        ;
        case H2_END          : return "h2"        ;
        case H3_START        : return "h3"        ;
        case H3_END          : return "h3"        ;
        case H4_START        : return "h4"        ;
        case H4_END          : return "h4"        ;
        case H5_START        : return "h5"        ;
        case H5_END          : return "h5"        ;
        case H6_START        : return "h6"        ;
        case H6_END          : return "h6"        ;
        case UL_START        : return "ul"        ;
        case UL_END          : return "ul"        ;
        case OL_START        : return "ol"        ;
        case OL_END          : return "ol"        ;
        case DIR_START       : return "dir"       ;
        case DIR_END         : return "dir"       ;
        case MENU_START      : return "menu"      ;
        case MENU_END        : return "menu"      ;
        case LI_START        : return "li"        ;
        case LI_END          : return "li"        ;
        case DL_START        : return "dl"        ;
        case DL_END          : return "dl"        ;
        case DT_START        : return "dt"        ;
        case DT_END          : return "dt"        ;
        case DD_START        : return "dd"        ;
        case DD_END          : return "dd"        ;
        case PRE_START       : return "pre"       ;
        case PRE_END         : return "pre"       ;
        case DIV_START       : return "div"       ;
        case DIV_END         : return "div"       ;
        case BLOCKQUOTE_START: return "blockquote";
        case BLOCKQUOTE_END  : return "blockquote";
        case HR              : return "hr"        ;
        case ADDRESS_START   : return "address"   ;
        case ADDRESS_END     : return "address"   ;
        case CENTER_START    : return "center"    ;
        case CENTER_END      : return "center"    ;
        case X_BLOCK_START   : return "_block"    ;
        case X_BLOCK_END     : return "_block"    ;
        // inline
        case A_START         : return "a"         ;
        case A_END           : return "a"         ;
        case IMG             : return "img"       ;
        case BR              : return "br"        ;
        case SPAN_START      : return "span"      ;
        case SPAN_END        : return "span"      ;
        case EM_START        : return "em"        ;
        case EM_END          : return "em"        ;
        case STRONG_START    : return "strong"    ;
        case STRONG_END      : return "strong"    ;
        case DFN_START       : return "dfn"       ;
        case DFN_END         : return "dfn"       ;
        case CODE_START      : return "code"      ;
        case CODE_END        : return "code"      ;
        case Q_START         : return "q"         ;
        case Q_END           : return "q"         ;
        case SUB_START       : return "sub"       ;
        case SUB_END         : return "sub"       ;
        case SUP_START       : return "sup"       ;
        case SUP_END         : return "sup"       ;
        case SAMP_START      : return "samp"      ;
        case SAMP_END        : return "samp"      ;
        case KBD_START       : return "kbd"       ;
        case KBD_END         : return "kbd"       ;
        case VAR_START       : return "var"       ;
        case VAR_END         : return "var"       ;
        case CITE_START      : return "cite"      ;
        case CITE_END        : return "cite"      ;
        case ABBR_START      : return "abbr"      ;
        case ABBR_END        : return "abbr"      ;
        case ACRONYM_START   : return "acronym"   ;
        case ACRONYM_END     : return "acronym"   ;
        case TT_START        : return "tt"        ;
        case TT_END          : return "tt"        ;
        case I_START         : return "i"         ;
        case I_END           : return "i"         ;
        case B_START         : return "b"         ;
        case B_END           : return "b"         ;
        case BIG_START       : return "big"       ;
        case BIG_END         : return "big"       ;
        case SMALL_START     : return "small"     ;
        case SMALL_END       : return "small"     ;
        case U_START         : return "u"         ;
        case U_END           : return "u"         ;
        case S_START         : return "s"         ;
        case S_END           : return "s"         ;
        case STRIKE_START    : return "strike"    ;
        case STRIKE_END      : return "strike"    ;
        case FONT_START      : return "font"      ;
        case FONT_END        : return "font"      ;
        case NOSCRIPT_START  : return "noscript"  ;
        case NOSCRIPT_END    : return "noscript"  ;
        // ruby
        case RUBY_START      : return "ruby"      ;
        case RUBY_END        : return "ruby"      ;
        case RB_START        : return "rb"        ;
        case RB_END          : return "rb"        ;
        case RP_START        : return "rp"        ;
        case RP_END          : return "rp"        ;
        case RT_START        : return "rt"        ;
        case RT_END          : return "rt"        ;
        // table
        case TABLE_START     : return "table"     ;
        case TABLE_END       : return "table"     ;
        case CAPTION_START   : return "caption"   ;
        case CAPTION_END     : return "caption"   ;
        case THEAD_START     : return "thead"     ;
        case THEAD_END       : return "thead"     ;
        case TBODY_START     : return "tbody"     ;
        case TBODY_END       : return "tbody"     ;
        case TFOOT_START     : return "tfoot"     ;
        case TFOOT_END       : return "tfoot"     ;
        case TR_START        : return "tr"        ;
        case TR_END          : return "tr"        ;
        case TH_START        : return "th"        ;
        case TH_END          : return "th"        ;
        case TD_START        : return "td"        ;
        case TD_END          : return "td"        ;
        // table
        case FORM_START      : return "form"      ;
        case FORM_END        : return "form"      ;
        case INPUT           : return "input"     ;
        case SELECT_START    : return "select"    ;
        case SELECT_END      : return "select"    ;
        case OPTGROUP_START  : return "optgroup"  ;
        case OPTGROUP_END    : return "optgroup"  ;
        case OPTION_START    : return "option"    ;
        case OPTION_END      : return "option"    ;
        case TEXTAREA_START  : return "textarea"  ;
        case TEXTAREA_END    : return "textarea"  ;
        case BUTTON_START    : return "button"    ;
        case BUTTON_END      : return "button"    ;
        case FIELDSET_START  : return "fieldset"  ;
        case FIELDSET_END    : return "fieldset"  ;
        case LEGEND_START    : return "legend"    ;
        case LEGEND_END      : return "legend"    ;
        case LABEL_START     : return "label"     ;
        case LABEL_END       : return "label"     ;
        default: // AVOID
        }

        return null;
    }

    /**
     * ꤵ줿ʸ󤫤顢פ֤ޤ
     *
     * @param  name ʸ
     *
     * @return ס
     *         ʾ <code>UNKNOWN</code>
     */
    public static int getType(String name) {
        Integer i = (Integer) TOKEN_NAMES.get(name.toUpperCase());

        if (i != null) {
            return i.intValue();
        }

        return UNKNOWN;
    }

    /** Ū */
    static {
        int[] typeList = {
            HTML_START      ,
            HTML_END        ,
            // header
            HEAD_START      ,
            HEAD_END        ,

            TITLE_START     ,
            TITLE_END       ,
            BASE            ,
            SCRIPT_START    ,
            SCRIPT_END      ,
            STYLE_START     ,
            STYLE_END       ,
            META            ,
            LINK            ,
            // frames
            FRAMESET_START  ,
            FRAMESET_END    ,
            FRAME           ,
            NOFRAMES_START  ,
            NOFRAMES_END    ,
            // body
            BODY_START      ,
            BODY_END        ,
            // block
            P_START         ,
            P_END           ,
            H1_START        ,
            H1_END          ,
            H2_START        ,
            H2_END          ,
            H3_START        ,
            H3_END          ,
            H4_START        ,
            H4_END          ,
            H5_START        ,
            H5_END          ,
            H6_START        ,
            H6_END          ,
            UL_START        ,
            UL_END          ,
            OL_START        ,
            OL_END          ,
            DIR_START       ,
            DIR_END         ,
            MENU_START      ,
            MENU_END        ,
            LI_START        ,
            LI_END          ,
            DL_START        ,
            DL_END          ,
            DT_START        ,
            DT_END          ,
            DD_START        ,
            DD_END          ,
            PRE_START       ,
            PRE_END         ,
            DIV_START       ,
            DIV_END         ,
            BLOCKQUOTE_START,
            BLOCKQUOTE_END  ,
            HR              ,
            ADDRESS_START   ,
            ADDRESS_END     ,
            CENTER_START    ,
            CENTER_END      ,
            //X_BLOCK_START   ,
            //X_BLOCK_END     ,
            // inline
            A_START         ,
            A_END           ,
            IMG             ,
            BR              ,
            SPAN_START      ,
            SPAN_END        ,
            EM_START        ,
            EM_END          ,
            STRONG_START    ,
            STRONG_END      ,
            DFN_START       ,
            DFN_END         ,
            CODE_START      ,
            CODE_END        ,
            Q_START         ,
            Q_END           ,
            SUB_START       ,
            SUB_END         ,
            SUP_START       ,
            SUP_END         ,
            SAMP_START      ,
            SAMP_END        ,
            KBD_START       ,
            KBD_END         ,
            VAR_START       ,
            VAR_END         ,
            CITE_START      ,
            CITE_END        ,
            ABBR_START      ,
            ABBR_END        ,
            ACRONYM_START   ,
            ACRONYM_END     ,
            TT_START        ,
            TT_END          ,
            I_START         ,
            I_END           ,
            B_START         ,
            B_END           ,
            BIG_START       ,
            BIG_END         ,
            SMALL_START     ,
            SMALL_END       ,
            U_START         ,
            U_END           ,
            S_START         ,
            S_END           ,
            STRIKE_START    ,
            STRIKE_END      ,
            FONT_START      ,
            FONT_END        ,
            NOSCRIPT_START  ,
            NOSCRIPT_END    ,
            // ruby
            RUBY_START      ,
            RUBY_END        ,
            RB_START        ,
            RB_END          ,
            RP_START        ,
            RP_END          ,
            RT_START        ,
            RT_END          ,
            // table
            TABLE_START     ,
            TABLE_END       ,
            CAPTION_START   ,
            CAPTION_END     ,
            THEAD_START     ,
            THEAD_END       ,
            TBODY_START     ,
            TBODY_END       ,
            TFOOT_START     ,
            TFOOT_END       ,
            TR_START        ,
            TR_END          ,
            TH_START        ,
            TH_END          ,
            TD_START        ,
            TD_END          ,
            // form
            FORM_START      ,
            FORM_END        ,
            INPUT           ,
            SELECT_START    ,
            SELECT_END      ,
            OPTGROUP_START  ,
            OPTGROUP_END    ,
            OPTION_START    ,
            OPTION_END      ,
            TEXTAREA_START  ,
            TEXTAREA_END    ,
            BUTTON_START    ,
            BUTTON_END      ,
            FIELDSET_START  ,
            FIELDSET_END    ,
            LEGEND_START    ,
            LEGEND_END      ,
            LABEL_START     ,
            LABEL_END       ,
        };
        String[] nameList = {
            "HTML_START"      ,
            "HTML_END"        ,
            // header
            "HEAD_START"      ,
            "HEAD_END"        ,

            "TITLE_START"     ,
            "TITLE_END"       ,
            "BASE_START"      ,
            "SCRIPT_START"    ,
            "SCRIPT_END"      ,
            "STYLE_START"     ,
            "STYLE_END"       ,
            "META_START"      ,
            "LINK_START"      ,
            // frame
            "FRAMESET_START"  ,
            "FRAMESET_END"    ,
            "FRAME_START"     ,
            "NOFRAMES_START"  ,
            "NOFRAMES_END"    ,
            // body
            "BODY_START"      ,
            "BODY_END"        ,
            // block
            "P_START"         ,
            "P_END"           ,
            "H1_START"        ,
            "H1_END"          ,
            "H2_START"        ,
            "H2_END"          ,
            "H3_START"        ,
            "H3_END"          ,
            "H4_START"        ,
            "H4_END"          ,
            "H5_START"        ,
            "H5_END"          ,
            "H6_START"        ,
            "H6_END"          ,
            "UL_START"        ,
            "UL_END"          ,
            "OL_START"        ,
            "OL_END"          ,
            "DIR_START"       ,
            "DIR_END"         ,
            "MENU_START"      ,
            "MENU_END"        ,
            "LI_START"        ,
            "LI_END"          ,
            "DL_START"        ,
            "DL_END"          ,
            "DT_START"        ,
            "DT_END"          ,
            "DD_START"        ,
            "DD_END"          ,
            "PRE_START"       ,
            "PRE_END"         ,
            "DIV_START"       ,
            "DIV_END"         ,
            "BLOCKQUOTE_START",
            "BLOCKQUOTE_END"  ,
            "HR_START"        ,
            "ADDRESS_START"   ,
            "ADDRESS_END"     ,
            "CENTER_START"    ,
            "CENTER_END"      ,
            //"X_BLOCK_START"   ,
            //"X_BLOCK_START"   ,
            // inline
            "A_START"         ,
            "A_END"           ,
            "IMG_START"       ,
            "BR_START"        ,
            "SPAN_START"      ,
            "SPAN_END"        ,
            "EM_START"        ,
            "EM_END"          ,
            "STRONG_START"    ,
            "STRONG_END"      ,
            "DFN_START"       ,
            "DFN_END"         ,
            "CODE_START"      ,
            "CODE_END"        ,
            "Q_START"         ,
            "Q_END"           ,
            "SUB_START"       ,
            "SUB_END"         ,
            "SUP_START"       ,
            "SUP_END"         ,
            "SAMP_START"      ,
            "SAMP_END"        ,
            "KBD_START"       ,
            "KBD_END"         ,
            "VAR_START"       ,
            "VAR_END"         ,
            "CITE_START"      ,
            "CITE_END"        ,
            "ABBR_START"      ,
            "ABBR_END"        ,
            "ACRONYM_START"   ,
            "ACRONYM_END"     ,
            "TT_START"        ,
            "TT_END"          ,
            "I_START"         ,
            "I_END"           ,
            "B_START"         ,
            "B_END"           ,
            "BIG_START"       ,
            "BIG_END"         ,
            "SMALL_START"     ,
            "SMALL_END"       ,
            "U_START"         ,
            "U_END"           ,
            "S_START"         ,
            "S_END"           ,
            "STRIKE_START"    ,
            "STRIKE_END"      ,
            "FONT_START"      ,
            "FONT_END"        ,
            "NOSCRIPT_START"  ,
            "NOSCRIPT_END"    ,
            // ruby
            "RUBY_START"      ,
            "RUBY_END"        ,
            "RB_START"        ,
            "RB_END"          ,
            "RP_START"        ,
            "RP_END"          ,
            "RT_START"        ,
            "RT_END"          ,
            // table
            "TABLE_START"     ,
            "TABLE_END"       ,
            "CAPTION_START"   ,
            "CAPTION_END"     ,
            "THEAD_START"     ,
            "THEAD_END"       ,
            "TBODY_START"     ,
            "TBODY_END"       ,
            "TFOOT_START"     ,
            "TFOOT_END"       ,
            "TR_START"        ,
            "TR_END"          ,
            "TH_START"        ,
            "TH_END"          ,
            "TD_START"        ,
            "TD_END"          ,
            // form
            "FORM_START"      ,
            "FORM_END"        ,
            "INPUT_START"     ,
            "SELECT_START"    ,
            "SELECT_END"      ,
            "OPTGROUP_START"  ,
            "OPTGROUP_END"    ,
            "OPTION_START"    ,
            "OPTION_END"      ,
            "TEXTAREA_START"  ,
            "TEXTAREA_END"    ,
            "BUTTON_START"    ,
            "BUTTON_END"      ,
            "FIELDSET_START"  ,
            "FIELDSET_END"    ,
            "LEGEND_START"    ,
            "LEGEND_END"      ,
            "LABEL_START"     ,
            "LABEL_END"       ,
        };

        if (typeList.length != nameList.length) {
            throw new IllegalArgumentException("typelist != namelist");
        }

        for (int i = 0; i < typeList.length; i++) {
            TOKEN_NAMES.put(nameList[i], new Integer(typeList[i]));
        }
    }
}
