/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.token.style;

import net.hizlab.kagetaka.addin.StyleManager;
import net.hizlab.kagetaka.rendering.Content;
import net.hizlab.kagetaka.rendering.Render;
import net.hizlab.kagetaka.rendering.Reporter;
import net.hizlab.kagetaka.token.StartToken;
import net.hizlab.kagetaka.token.TokenTypes;

/**
 * ϤΥȡγϤɽ륹ѡ饹Ǥ
 *
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.1.1.1 $
 */
public class StyleStartToken extends StartToken {
    private boolean isIncludedStyle;
    /**
     * ϤΥȡγϤޤ
     *
     * @param  content  ƥ
     * @param  line     ֹ (<code>1</code> )
     *                  Ǥʤ <code>0</code>
     * @param  column    (<code>1</code> )
     *                  Ǥʤ <code>0</code>
     * @param  reporter 顼ݡ
     * @param  type     ȡγϤɽ
     * @param  complete 䴰ƺ줿 <code>true</code>
     *                  ʳξ <code>false</code>
     */
    public StyleStartToken(Content content, int line, int column, Reporter reporter, int type, boolean complete) {
        super(content, line, column, reporter, type, complete);
    }

    /** {@inheritDoc} */
    protected boolean setAttributeImple(String key, String value)
            throws IllegalArgumentException {
        if (attribute == null) {
            attribute = new StyleAttribute(content, lineNumber, columnNumber, reporter);
        }

        return attribute.setAttribute(key, value);
    }

    /** {@inheritDoc} */
    protected StartToken getDefaultParentTokenImple() {
        int tokenType = TokenTypes.UNKNOWN;

        switch (type) {
        case TokenTypes.STYLE_START: tokenType = TokenTypes.HEAD_START; break;
        default:
            return null;
        }

        return new StartToken(content, lineNumber, columnNumber, reporter, tokenType, true);
    }

    /** {@inheritDoc} */
    protected boolean isContentsImple(int childToken) {
        switch (type) {
        case TokenTypes.STYLE_START:
            return true;
        default: // AVOID
        }

        return false;
    }

    /** {@inheritDoc} */
    protected void renderImple(Render render) {
        switch (type) {
        case TokenTypes.STYLE_START:
            if (content != null && !isIncludedStyle) {
                isIncludedStyle = true;

                String data = rawContent.trim();
                if (data.startsWith("<!--")) {
                    data = data.substring(4);
                }
                if (data.endsWith("-->")) {
                    data = data.substring(0, data.length() - 3);
                }

                StyleManager sm = render.getStyleManager(true);
                if (sm != null) {
                    sm.addStyle(content, data);
                }
            }
            break;
        default: // AVOID
        }
    }
}
